# Standard Library
import logging
import re
from typing import List, Optional

# Third Party
import tomlkit
from fastapi import APIRouter, Depends, File, HTTPException, Query, UploadFile, status
from fastapi.responses import FileResponse
from fastapi_cache.decorator import cache
from packaging.version import Version
from pydantic import Required
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    CACHE_NAMESPACE_FINDING,
    CACHE_NAMESPACE_RULE,
    CACHE_NAMESPACE_RULE_PACK,
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    ERROR_MESSAGE_500,
    ERROR_MESSAGE_503,
    REDIS_CACHE_EXPIRE,
    RULE_PACKS_TAG,
    RWS_ROUTE_RULE_PACKS,
)
from resc_backend.resc_web_service.cache_manager import CacheManager
from resc_backend.resc_web_service.crud import rule as rule_crud
from resc_backend.resc_web_service.crud import rule_pack as rule_pack_crud
from resc_backend.resc_web_service.crud import rule_tag as rule_tag_crud
from resc_backend.resc_web_service.dependencies import get_db_connection
from resc_backend.resc_web_service.helpers.resc_swagger_models import (
    Model400,
    Model404,
    Model409,
    Model422,
)
from resc_backend.resc_web_service.helpers.rule import (
    create_toml_dictionary,
    create_toml_rule_file,
    get_mapped_global_allow_list_obj,
    map_dictionary_to_rule_allow_list_object,
    validate_uploaded_file_and_read_content,
)
from resc_backend.resc_web_service.schema.pagination_model import PaginationModel
from resc_backend.resc_web_service.schema.rule import RuleCreate
from resc_backend.resc_web_service.schema.rule_allow_list import RuleAllowList
from resc_backend.resc_web_service.schema.rule_pack import RulePackCreate, RulePackRead

router = APIRouter(prefix=f"{RWS_ROUTE_RULE_PACKS}", tags=[RULE_PACKS_TAG])

logger = logging.getLogger(__name__)


@router.get(
    "/versions",
    response_model=PaginationModel[RulePackRead],
    summary="Get rule packs",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the rule-packs"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE_PACK, expire=REDIS_CACHE_EXPIRE)
def get_rule_packs(
    version: Optional[str] = Query(None, pattern=r"^\d+(?:\.\d+){2}$"),
    active: Optional[bool] = Query(None, description="Filter on active rule packs"),
    skip: int = Query(default=0, ge=0),
    limit: int = Query(default=DEFAULT_RECORDS_PER_PAGE_LIMIT, ge=1),
    db_connection: Session = Depends(get_db_connection),
) -> PaginationModel[RulePackRead]:
    """
        Retrieve rule packs

    - **db_connection**: Session of the database connection
    - **version**: Optional, filter on rule pack version
    - **active**: Optional, filter on active rule pack
    - **skip**: Integer amount of records to skip, to support pagination
    - **limit**: Integer amount of records to return, to support pagination
    - **return**: [RulePackRead]
        The output will contain a PaginationModel containing the list of RulePackRead type objects,
        or an empty list if no rule pack was found
    """
    rule_packs = rule_pack_crud.get_rule_packs(
        db_connection=db_connection,
        version=version,
        active=active,
        skip=skip,
        limit=limit,
    )
    total_rule_packs_count = rule_pack_crud.get_total_rule_packs_count(
        db_connection=db_connection, version=version, active=active
    )
    return PaginationModel[RulePackRead](data=rule_packs, total=total_rule_packs_count, limit=limit, skip=skip)


@router.get(
    "",
    summary="Download rule pack in TOML format",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Download the rule-pack in TOML format"},
        404: {
            "model": Model404,
            "description": "No rule-pack of version <version_id> found",
        },
        422: {
            "model": Model422,
            "description": "Version <version_id> is not a valid semantic version",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def download_rule_pack_toml_file(
    version: Optional[str] = Query(None, pattern=r"^\d+(?:\.\d+){2}$"),
    db_connection: Session = Depends(get_db_connection),
) -> FileResponse:
    """
        Download rule pack in TOML format

    - **db_connection**: Session of the database connection
    - **version**: Optional, filter on rule pack version
    - **return**: [FileResponse] The output returns rule pack file downloaded in TOML format
    """
    if not version:
        logger.info("rule pack version not specified, downloading the currently active version")
    rule_pack_from_db = read_rule_pack(version=version, db_connection=db_connection)
    if rule_pack_from_db:
        version = rule_pack_from_db.version
        rules = rule_crud.get_rules_by_rule_pack_version(db_connection=db_connection, rule_pack_version=version)
        rule_tag_names = rule_tag_crud.get_rule_tag_names_by_rule_pack_version(
            db_connection=db_connection, rule_pack_version=version
        )
        global_allow_list = rule_crud.get_global_allow_list_by_rule_pack_version(
            db_connection=db_connection, rule_pack_version=version
        )
        generated_toml_dict = create_toml_dictionary(version, rules, global_allow_list, rule_tag_names)
    else:
        raise HTTPException(status_code=404, detail=f"No rule pack found with version {version}")

    toml_file = create_toml_rule_file(generated_toml_dict)
    return FileResponse(toml_file.name, filename="RESC-SECRETS-RULE.toml")


@router.post(
    "",
    summary="Upload rule pack in TOML format",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Upload the rule-pack in TOML format"},
        400: {
            "model": Model400,
            "description": "No properties defined for rule allow list",
        },
        409: {
            "model": Model409,
            "description": "Rule-pack version <version_id> already exists",
        },
        422: {
            "model": Model422,
            "description": "Version <version_id> is not a valid semantic version",
        },
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
async def upload_rule_pack_toml_file(
    version: str = Query(default=Required, pattern=r"^\d+(?:\.\d+){2}$"),
    rule_file: UploadFile = File(...),
    db_connection: Session = Depends(get_db_connection),
) -> dict:
    """
        Upload TOML rule pack to database

    - **db_connection**: Session of the database connection
    - **version**: Version of the rule pack to be uploaded
    - **rule_file**: TOML rule pack file to be uploaded
    - **return**: dict The output returns uploaded rule pack name in dictionary format
    """
    content = validate_uploaded_file_and_read_content(rule_file)
    toml_rule_dictionary = tomlkit.loads(content)

    # Check if rule pack version exists
    rule_pack_from_db = read_rule_pack(version=version, db_connection=db_connection)
    if rule_pack_from_db:
        raise HTTPException(
            status_code=409,
            detail=f"Unable to process rules. Rule pack version " f"{version} already exists",
        )

    # Insert in to RULE_ALLOW_LIST for storing global allow list
    created_global_rule_allow_list = None
    global_allow_list: RuleAllowList = get_mapped_global_allow_list_obj(toml_rule_dictionary)
    if global_allow_list:
        created_global_rule_allow_list = create_rule_allow_list(
            rule_allow_list=global_allow_list, db_connection=db_connection
        )
        if not created_global_rule_allow_list.id_:
            logger.warning("Creating global rule allow list failed with an error")

    # # Insert in to RULE_PACK
    global_allow_list_id = (
        created_global_rule_allow_list.id_
        if created_global_rule_allow_list and created_global_rule_allow_list.id_
        else None
    )

    # Determine if uploaded rule pack needs to be activated
    current_newest_rule_pack = rule_pack_crud.get_newest_rule_pack(db_connection)
    activate_uploaded_rule_pack = determine_uploaded_rule_pack_activation(version, current_newest_rule_pack)

    rule_pack = RulePackCreate(
        version=version,
        active=activate_uploaded_rule_pack,
        global_allow_list=global_allow_list_id,
    )
    created_rule_pack_version = create_rule_pack_version(rule_pack=rule_pack, db_connection=db_connection)
    if created_rule_pack_version.version and activate_uploaded_rule_pack:
        # Update older rule packs to inactive
        rule_pack_crud.make_older_rule_packs_to_inactive(latest_rule_pack_version=version, db_connection=db_connection)
    else:
        logger.warning("Creating rule pack failed with an error")

    # Insert in to RULES
    if created_rule_pack_version and created_rule_pack_version.version:
        insert_rules(
            version=version,
            toml_rule_dictionary=toml_rule_dictionary,
            db_connection=db_connection,
        )

        # Clear cache related to rule-pack
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE)
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_RULE_PACK)
        await CacheManager.clear_cache_by_namespace(namespace=CACHE_NAMESPACE_FINDING)
    return {"filename": rule_file.filename}


def read_rule_pack(version: Optional[str] = None, db_connection: Session = Depends(get_db_connection)) -> RulePackRead:
    """
        Read active rule pack from database
    :param version:
        optional, version of the rule pack to be fetched else latest rule pack version will be fetched
    :param db_connection:
        Session of the database connection
    :return: RulePackRead
        The output returns RulePackRead type object
    """
    if version:
        regex = re.compile(r"^\d+(?:\.\d+){2}$")
        if not re.fullmatch(regex, version):
            raise HTTPException(
                status_code=422,
                detail=f"Version {version} is not a valid semantic version",
            )
    db_rule_pack = rule_pack_crud.get_rule_pack(db_connection=db_connection, version=version)
    return db_rule_pack


def create_rule_pack_version(rule_pack: RulePackCreate, db_connection: Session = Depends(get_db_connection)):
    """
        Create rule pack version in database
    :param rule_pack:
        RulePackCreate object to be created
    :param db_connection:
        Session of the database connection
    """
    return rule_pack_crud.create_rule_pack_version(db_connection=db_connection, rule_pack=rule_pack)


def create_rule_allow_list(rule_allow_list: RuleAllowList, db_connection: Session = Depends(get_db_connection)):
    """
        Create rule allow list in database
    :param rule_allow_list:
        RuleAllowList object to be created
    :param db_connection:
        Session of the database connection
    """
    if (
        rule_allow_list.paths
        or rule_allow_list.commits
        or rule_allow_list.stop_words
        or rule_allow_list.description
        or rule_allow_list.regexes
    ):
        return rule_crud.create_rule_allow_list(db_connection=db_connection, rule_allow_list=rule_allow_list)
    raise HTTPException(status_code=400, detail="No properties defined for rule allow list")


def insert_rules(
    version: str,
    toml_rule_dictionary: dict,
    db_connection: Session = Depends(get_db_connection),
):
    """
        Create rules in database
    :param version:
        version of the rule pack
    :param toml_rule_dictionary:
        rule pack toml in dictionary format
    :param db_connection:
        Session of the database connection
    """
    if "rules" in toml_rule_dictionary:
        rule_list = toml_rule_dictionary.get("rules")
        for rule in rule_list:
            keywords = None
            rule_name = rule["id"] if "id" in rule else None
            description = rule["description"] if "description" in rule else None
            regex = rule["regex"] if "regex" in rule else None
            entropy = rule["entropy"] if "entropy" in rule else None
            secret_group = rule["secretGroup"] if "secretGroup" in rule else None
            path = rule["path"] if "path" in rule else None

            if "keywords" in rule:
                keyword_array = rule["keywords"]
                keywords = ",".join(keyword_array)

            # Insert in to RULE_ALLOW_LIST for storing individual rule specific  allow list
            allow_list = rule["allowlist"] if "allowlist" in rule else None
            rule_allow_list_obj = map_dictionary_to_rule_allow_list_object(allow_list)
            created_rule_allow_list = None
            if rule_allow_list_obj:
                created_rule_allow_list = create_rule_allow_list(
                    rule_allow_list=rule_allow_list_obj, db_connection=db_connection
                )

            # Insert in to RULES
            if allow_list and created_rule_allow_list and created_rule_allow_list.id_:
                created_allow_list_id = created_rule_allow_list.id_
            else:
                created_allow_list_id = None
            rule_obj = RuleCreate(
                rule_pack=version,
                allow_list=created_allow_list_id,
                rule_name=rule_name,
                description=description,
                entropy=entropy,
                secret_group=secret_group,
                regex=regex,
                path=path,
                keywords=keywords,
            )
            created_rule = rule_crud.create_rule(rule=rule_obj, db_connection=db_connection)
            if not created_rule.id_:
                logger.warning(f"Creating rule failed for Rule: {rule_name}")
            if "tags" in rule:
                _ = rule_tag_crud.create_rule_tag(
                    db_connection=db_connection,
                    rule_id=created_rule.id_,
                    tags=rule["tags"],
                )


@router.get(
    "/tags",
    response_model=List[str],
    summary="Get rule packs' tags",
    status_code=status.HTTP_200_OK,
    responses={
        200: {"description": "Retrieve all the tags related to a rule-pack[s]"},
        500: {"description": ERROR_MESSAGE_500},
        503: {"description": ERROR_MESSAGE_503},
    },
)
@cache(namespace=CACHE_NAMESPACE_RULE_PACK, expire=REDIS_CACHE_EXPIRE)
def get_rule_packs_tags(
    versions: Optional[List[str]] = Query(None, alias="version", title="version"),
    db_connection: Session = Depends(get_db_connection),
) -> List[str]:
    """
        Retrieve rule pack related tags

    :param db_connection: Session of the database connection
    :param versions: Optional, filter on rule pack version, if not provided filter on active.
    :return: List[str]
        The output will contain a list of tags related to one or more rule-packs.
    """
    if not versions:
        active_rule_pack = rule_pack_crud.get_current_active_rule_pack(db_connection=db_connection)
        if not active_rule_pack:
            raise HTTPException(status_code=500, detail="No currently active rule pack.")
        versions = [active_rule_pack.version]
    rule_packs_tags = rule_pack_crud.get_rule_packs_tags(db_connection=db_connection, versions=versions)
    return rule_packs_tags


def determine_uploaded_rule_pack_activation(
    requested_rule_pack_version: str, latest_rule_pack_from_db: RulePackRead
) -> bool:
    """
        Determine if rule pack needs to be activated
    :param requested_rule_pack_version:
        version of the rule pack uploaded
    :param latest_rule_pack_from_db:
        latest rule pack in RulePackRead object
    :return: boolean
        The output returns true if rule pack needs to be activated else returns false
    """
    if latest_rule_pack_from_db:
        if Version(latest_rule_pack_from_db.version) < Version(requested_rule_pack_version):
            logger.info(
                f"Uploaded rule pack is of version '{requested_rule_pack_version}', using it to replace "
                f"'{latest_rule_pack_from_db.version}' as the active one."
            )
            activate_uploaded_rule_pack = True
        else:
            if not latest_rule_pack_from_db.active:
                logger.info(
                    f"There is already a more recent rule pack present in the database "
                    f"'{latest_rule_pack_from_db.version}', but it is set to inactive. "
                    f"Activating the uploaded rule pack '{requested_rule_pack_version}'"
                )
                activate_uploaded_rule_pack = True
            else:
                logger.info(
                    f"Uploaded rule pack is of version '{requested_rule_pack_version}', the existing rule pack "
                    f"'{latest_rule_pack_from_db.version}' is kept as the active one."
                )
                activate_uploaded_rule_pack = False
    else:
        logger.info(
            f"No existing rule pack found, So activating the uploaded rule pack '{requested_rule_pack_version}'"
        )
        activate_uploaded_rule_pack = True
    return activate_uploaded_rule_pack



# Standard Library
from typing import List

# Third Party
from sqlalchemy import func
from sqlalchemy.orm import Session

# First Party
from resc_backend.constants import (
    DEFAULT_RECORDS_PER_PAGE_LIMIT,
    MAX_RECORDS_PER_PAGE_LIMIT,
)
from resc_backend.db.model import DBVcsInstance
from resc_backend.resc_web_service.crud import finding as finding_crud
from resc_backend.resc_web_service.crud import repository as repository_crud
from resc_backend.resc_web_service.crud import scan as scan_crud
from resc_backend.resc_web_service.crud import scan_finding as scan_finding_crud
from resc_backend.resc_web_service.schema import vcs_instance as vcs_instance_schema
from resc_backend.resc_web_service.schema.vcs_provider import VCSProviders


def get_vcs_instance(db_connection: Session, vcs_instance_id: int):
    vcs_instance = db_connection.query(DBVcsInstance).filter(DBVcsInstance.id_ == vcs_instance_id).first()
    return vcs_instance


def get_vcs_instances(
    db_connection: Session,
    skip: int = 0,
    limit: int = DEFAULT_RECORDS_PER_PAGE_LIMIT,
    vcs_provider_type: VCSProviders = None,
    vcs_instance_name: str = None,
) -> List[DBVcsInstance]:
    """
        Retrieve all vcs_instances records
    :param db_connection:
        Session of the database connection
    :param vcs_provider_type:
        optional filtering by VCS Provider type
    :param vcs_instance_name:
        optional filtering by VCS Provider name
    :param skip:
        integer amount of records to skip to support pagination
    :param limit:
        integer amount of records to return, to support pagination
    :return: [DBVcsInstance]
        List of DBVcsInstance objects
    """
    limit_val = MAX_RECORDS_PER_PAGE_LIMIT if limit > MAX_RECORDS_PER_PAGE_LIMIT else limit
    query = db_connection.query(DBVcsInstance)

    if vcs_provider_type:
        query = query.filter(DBVcsInstance.provider_type == vcs_provider_type)

    if vcs_instance_name:
        query = query.filter(DBVcsInstance.name == vcs_instance_name)

    vcs_instances = query.order_by(DBVcsInstance.id_).offset(skip).limit(limit_val).all()
    return vcs_instances


def get_vcs_instances_count(
    db_connection: Session,
    vcs_provider_type: VCSProviders = None,
    vcs_instance_name: str = None,
) -> int:
    """
        Retrieve count of vcs_instances records optionally filtered by VCS provider
    :param db_connection:
        Session of the database connection
    :param vcs_provider_type:
        optional vcs_provider_type filtering the vcs provider type for which to count vcs instances
    :param vcs_instance_name:
        optional vcs_instance_name filtering the vcs provider name for which to count vcs instances
    :return: total_count
        count of vcs instances
    """
    query = db_connection.query(func.count(DBVcsInstance.id_))

    if vcs_provider_type:
        query = query.filter(DBVcsInstance.provider_type == vcs_provider_type)

    if vcs_instance_name:
        query = query.filter(DBVcsInstance.name == vcs_instance_name)

    total_count = query.scalar()
    return total_count


def update_vcs_instance(
    db_connection: Session,
    vcs_instance_id: int,
    vcs_instance: vcs_instance_schema.VCSInstanceCreate,
) -> DBVcsInstance:
    db_vcs_instance: DBVcsInstance = db_connection.query(DBVcsInstance).filter_by(id_=vcs_instance_id).first()
    db_vcs_instance.name = vcs_instance.name
    db_vcs_instance.provider_type = vcs_instance.provider_type
    db_vcs_instance.port = vcs_instance.port
    db_vcs_instance.scheme = vcs_instance.scheme
    db_vcs_instance.organization = vcs_instance.organization
    db_vcs_instance.hostname = vcs_instance.hostname
    db_vcs_instance.scope = ",".join(vcs_instance.scope)
    db_vcs_instance.exceptions = ",".join(vcs_instance.exceptions)

    db_connection.commit()
    db_connection.refresh(db_vcs_instance)
    return db_vcs_instance


def create_vcs_instance(db_connection: Session, vcs_instance: vcs_instance_schema.VCSInstanceCreate) -> DBVcsInstance:
    db_vcs_instance = DBVcsInstance(
        name=vcs_instance.name,
        provider_type=vcs_instance.provider_type,
        port=vcs_instance.port,
        scheme=vcs_instance.scheme,
        organization=vcs_instance.organization,
        hostname=vcs_instance.hostname,
        scope=",".join(vcs_instance.scope),
        exceptions=",".join(vcs_instance.exceptions),
    )
    db_connection.add(db_vcs_instance)
    db_connection.commit()
    db_connection.refresh(db_vcs_instance)
    return db_vcs_instance


def create_vcs_instance_if_not_exists(db_connection: Session, vcs_instance: vcs_instance_schema.VCSInstanceCreate):
    # Query the database to see if the vcs_instance object exists based on the unique constraint parameters
    db_select_vcs_instance = (
        db_connection.query(DBVcsInstance)
        .filter(
            DBVcsInstance.provider_type == vcs_instance.provider_type,
            DBVcsInstance.scheme == vcs_instance.scheme,
            DBVcsInstance.hostname == vcs_instance.hostname,
            DBVcsInstance.port == vcs_instance.port,
            DBVcsInstance.organization == vcs_instance.organization,
        )
        .first()
    )
    if db_select_vcs_instance is not None:
        return db_select_vcs_instance

    # Create non-existing vcs_instance object
    return create_vcs_instance(db_connection, vcs_instance)


def delete_vcs_instance(db_connection: Session, vcs_instance_id: int, delete_related: bool = False):
    """
        Delete a vcs instance object
    :param db_connection:
        Session of the database connection
    :param vcs_instance_id:
        id of the vcs instance to be deleted
    :param delete_related:
        if related records need to be deleted
    """
    if delete_related:
        scan_finding_crud.delete_scan_finding_by_vcs_instance_id(db_connection, vcs_instance_id=vcs_instance_id)
        finding_crud.delete_findings_by_vcs_instance_id(db_connection, vcs_instance_id=vcs_instance_id)
        scan_crud.delete_scans_by_vcs_instance_id(db_connection, vcs_instance_id=vcs_instance_id)
        repository_crud.delete_repositories_by_vcs_instance_id(db_connection, vcs_instance_id=vcs_instance_id)
    db_vcs_instance = db_connection.query(DBVcsInstance).filter_by(id_=vcs_instance_id).first()
    db_connection.delete(db_vcs_instance)
    db_connection.commit()



# Standard Library
import logging
import os
import urllib

# Third Party
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

# First Party
from resc_backend.db.engine_azure import create_azure_engine
from resc_backend.db.model import Base

basedir = os.path.abspath(os.path.dirname(__file__))
logger = logging.getLogger(__name__)

DB_CONNECTION_STRING = os.environ.get("DB_CONNECTION_STRING")
try:
    env_variables = {}
    for key, value in os.environ.items():
        if "pass" in key.lower():
            env_variables[key] = urllib.parse.quote(value)
        else:
            env_variables[key] = value
    DB_CONNECTION_STRING = DB_CONNECTION_STRING.format(**env_variables)
except AttributeError:
    logger.warning("Missing DB Connection environment variables, using SQLite database")
    DB_CONNECTION_STRING = ""
SQL_ALCHEMY_POOL_SIZE_DEFAULT = 25
SQL_ALCHEMY_MAX_OVERFLOW_DEFAULT = 15

echo_queries = os.getenv("SQL_ALCHEMY_ECHO_QUERIES", "False").lower() in (
    "true",
    "1",
    "y",
)
pool_size = int(os.environ.get("SQL_ALCHEMY_POOL_SIZE", SQL_ALCHEMY_POOL_SIZE_DEFAULT))
max_overflow = int(os.environ.get("SQL_ALCHEMY_MAX_OVERFLOW", SQL_ALCHEMY_MAX_OVERFLOW_DEFAULT))

if DB_CONNECTION_STRING:
    logger.info("Using provided environment variable to connect to the Database")
    use_azure_token_auth = os.environ.get("DB_USE_AZURE_TOKEN_AUTH", "False").lower() in ("true", "1", "y")
    if use_azure_token_auth:
        engine = create_azure_engine(
            connection_string=DB_CONNECTION_STRING,
            echo_queries=echo_queries,
            pool_size=pool_size,
            max_overflow=max_overflow,
        )
    else:
        engine = create_engine(
            DB_CONNECTION_STRING,
            echo=echo_queries,
            pool_size=pool_size,
            max_overflow=max_overflow,
        )
else:
    DATABASE_URL = "sqlite:///" + os.path.join(basedir, "db.sqlite?check_same_thread=False")
    logger.info(f"Database environment variables were not provided, defaulting to {DATABASE_URL}")
    engine = create_engine(DATABASE_URL, echo=echo_queries)
    Base.metadata.create_all(engine, checkfirst=True)

Session = sessionmaker()
